# This file is automatically generated by pyo3_stub_gen
# ruff: noqa: E501, F401

import builtins
from . import expression
from . import instructions
from . import program
from . import waveforms

class PickleError(builtins.QuilError):
    r"""
    Errors when trying to pickle or deepcopy.
    """
    ...

class QuilError(builtins.Exception):
    r"""
    Base exception type for errors raised by this package.
    """
    ...

class QuilValueError(builtins.QuilError):
    r"""
    Input value is invalid.
    """
    ...

class ToQuilError(builtins.QuilError):
    r"""
    Errors which can occur when converting a Quil item to a string.
    """
    ...

