//! This module contains analysis tools for Quil programs.

// Copyright 2024 Rigetti Computing
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
// http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

mod control_flow_graph;
mod qubit_graph;

pub use control_flow_graph::{
    BasicBlock, BasicBlockOwned, BasicBlockScheduleError, BasicBlockTerminator,
    BasicBlockTerminatorOwned, ControlFlowGraph, ControlFlowGraphOwned,
};
pub use qubit_graph::{QubitGraph, QubitGraphError};

#[cfg(test)]
pub(crate) mod test_programs;
