# This file is automatically generated by pyo3_stub_gen
# ruff: noqa: E501, F401

import builtins

class IdentifierValidationError(builtins.QuilError):
    r"""
    Errors that may occur when validating a Quil identifier.
    """
    ...

def validate_identifier(ident:builtins.str) -> None:
    r"""
    Returns an error if the given identifier is not a valid Quil Identifier.
    """

def validate_user_identifier(ident:builtins.str) -> None:
    r"""
    Returns an error if the given identifier is reserved, or if it is not a valid Quil identifier
    """

